@echo off
setlocal
set ROOT_DIR=%~dp0..\..\..\..
set VSROOT_DIR=%~dp0..\..
call "%ROOT_DIR%\node.exe" "%VSROOT_DIR%\out\server-cli.js" "code-server" "1.73.1" "6261075646f055b99068d3688932416f2346dd3b" "code-server.cmd" "--openExternal" %*
endlocal
